declare module "api" {
    /** A function that returns whether a value is valid. */
    export type Predicate = (value: any) => boolean;
    /**
     * Interface implemented by all validation rules.
     */
    export interface Rule {
        /** Unique identifier for this rule. Used to programmatically identify rules, e.g. for building localized messages. */
        readonly id: string;
        /** Returns an English description of this rule. */
        describe(): string;
        /**
         * Tests whether the given value is valid according to this rule.
         *
         * @param value the value to validate
         */
        test(value: any): boolean;
    }
    /**
     * Interface for instantiated validation rules, which may contain arbitrary configuration.
     */
    export interface RuleInstance extends Rule, Record<string, any> {
    }
    /**
     * Exception class thrown or returned when validation errors are encountered.
     */
    export class ValidationError extends Error {
        /**
         * The failing validation rule.
         */
        readonly rule: RuleInstance;
        /**
         * The name of the variable containing the invalid value.
         */
        readonly variable?: string;
        /**
         * Creates a new ValidationError with the given error message and failing rule.
         *
         * @param message the error message
         * @param rule the failing validation rule
         */
        constructor(message: string, rule: RuleInstance, variable?: string);
    }
    /**
     * Interface used to actually perform validation.
     */
    export interface Validator {
        /** Returns a new validator that allows null or undefined, in addition to valid values. */
        optional(): this;
        /** Returns a new validator that allows null, in addition to valid values. */
        orNull(): this;
        /** Returns a new validator that allows undefined, in addition to valid values. */
        orUndefined(): this;
        /** Returns an English description of the rules checked by this validator. */
        describe(): string;
        /**
         * Validates the given value and throws a ValidationError if it is invalid.
         *
         * @param value the value to validate
         * @param name the name of the variable containing the value
         */
        check(value: any, name?: string): void;
        /**
         * Tests whether the given value is valid according to the rules checked by this validator.
         *
         * @param value the value to validate
         */
        test(value: any): boolean;
        /**
         * Validates the given value and returns ValidationErrors for any and all failing rules.
         *
         * @param value the value to validate
         * @param name the name of the variable containing the value
         */
        testAll(value: any, name?: string): ValidationError[];
    }
    /**
     * Mix-in interface for building validators that can be inverted with `not`.
     */
    export interface InvertibleValidation {
        /**
         * Returns a validation builder that inverts/negates the next validation rule method.
         * Double inversion/negation (e.g. `.not.not`) is not allowed.
         */
        readonly not: Omit<InvertedValidation<this>, 'not'>;
    }
    /**
     * Mix-in interface for building validators that apply to any type.
     */
    export interface AnyValidation extends InvertibleValidation {
        /**
         * Checks that at least one of the given validators passes, essentially providing an `or` operator.
         * If no validators are provided, no values are considered valid.
         *
         * @param validators an array of validators to apply
         */
        anyOf(...validators: Validator[]): this & Validator;
        /**
         * Returns a new validator that checks that the validated value equals (using `==`) the given value.
         *
         * @param value the value is compare against using `==`
         */
        equal(value: any): this & Validator;
        /**
         * Returns a new validator that checks that the validated value strictly equals (using `===`) the given value.
         *
         * @param value the value is compare against using `===`
         */
        exact(value: any): this & Validator;
        /**
         * Returns a new validator that checks that the validated value against the given predicate function.
         *
         * @param test the predicate used to test each value
         * @param description the description of this predicate (used in error messages)
         */
        predicate(test: Predicate, description: string): this & Validator;
    }
    /**
     * Mix-in interface for building validators that apply to numeric values.
     * @typeparam V return type for each rule builder method
     */
    export interface NumericValidationReturning<V> extends AnyValidation {
        /**
         * Returns a new validator that checks that the validated value is numeric and within the given range, inclusive.
         * Throws an Error if `min > max`.
         *
         * @param min the minimum allowed value
         * @param max the maximum allowed value
         */
        between(min: number, max: number): V;
        /**
         * Returns a new validator that checks that the validated value is numeric and greater than the given minimum.
         *
         * @param bound the exclusive lower bound
         */
        greaterThan(bound: number): V;
        /**
         * Returns a new validator that checks that the validated value is numeric and greater than or equal to the given minimum.
         *
         * @param bound the inclusive lower bound
         */
        greaterThanOrEqual(bound: number): V;
        /**
         * Returns a new validator that checks that the validated value is numeric and less than the given minimum.
         *
         * @param bound the exclusive upper bound
         */
        lessThan(bound: number): V;
        /**
         * Returns a new validator that checks that the validated value is numeric and less than or equal to the given minimum.
         *
         * @param bound the inclusive upper bound
         */
        lessThanOrEqual(bound: number): V;
        /** Returns a new validator that checks that the validated value is numeric and positive (`> 0`). */
        positive(): V;
        /** Returns a new validator that checks that the validated value is numeric and negative (`< 0`). */
        negative(): V;
    }
    /**
     * Mix-in type for non-inverted validation builders that apply to numeric values.
     */
    export type NumericValidation = NumericValidationReturning<NumericValidation & Validator>;
    /**
     * Mix-in interface for building validators that apply to types with a length, such as strings and arrays.
     * @typeparam V return type for each rule builder method
     */
    export interface LengthValidation<V> extends InvertibleValidation {
        /**
         * Checks that the length of the validated value is within the given range, inclusive.
         * If `max` is omitted, it is assumed to equal `min`.
         * Throws an Error if `min > max`.
         *
         * @param min the minimum allowed length
         * @param max the maximum allowed length
         */
        length(min: number, max?: number): V;
        /**
         * Checks that the length of the validated value is greater than or equal to the given minimum.
         *
         * @param min the minimum allowed length
         */
        minLength(min: number): V;
        /**
         * Checks that the length of the validated value is less than or equal to the given maximum.
         *
         * @param max the maximum allowed length
         */
        maxLength(max: number): V;
        /** Returns a new validator that checks that the validated value is an empty string or array. */
        empty(): V;
    }
    /**
     * Mix-in interface for building validators that apply to strings.
     * @typeparam V return type for each rule builder method
     */
    export interface StringValidationReturning<V> extends AnyValidation, LengthValidation<V> {
        /**
         * Returns a new validator that checks that the validated value is a string matching the given regular expression.
         *
         * @param regex a regular expression to match
         */
        pattern(regex: RegExp): V;
        /**
         * Returns a new validator that checks that the validated value is a string containing the given substring.
         *
         * @param str a substring to search for
         */
        contains(str: string): V;
        /**
         * Returns a new validator that checks that the validated value is a string starting with the given substring.
         *
         * @param str a substring to match at the beginning of the value
         */
        startsWith(str: string): V;
        /**
         * Returns a new validator that checks that the validated value is a string ending with the given substring.
         *
         * @param str a substring to match at the end of the value
         */
        endsWith(str: string): V;
    }
    /**
     * Mix-in type for non-inverted validation builders that apply to strings.
     */
    export type StringValidation = StringValidationReturning<StringValidation & Validator>;
    /**
     * Mix-in interface for building validators that apply to arrays.
     * @typeparam V return type for each rule builder method
     */
    export interface ArrayValidationReturning<V> extends LengthValidation<V> {
        /**
         * Returns a new validator that checks that the validated value is an array containing the given element.
         *
         * @param value an element value to search the array for
         */
        includes(value: any): V;
        /**
         * Returns a new validator that checks that the validated value is an array for which every element passes the given validator.
         *
         * @param elementValidator a validator to apply to each element
         */
        every(elementValidator: Validator): V;
        /**
         * Returns a new validator that checks that the validated value is an array for which at least one element passes the given validator.
         *
         * @param elementValidator a validator to apply to each element
         */
        some(elementValidator: Validator): V;
    }
    /**
     * Mix-in type for non-inverted validation builders that apply to arrays.
     */
    export type ArrayValidation = ArrayValidationReturning<ArrayValidation & Validator>;
    /**
     * Mix-in interface for building validators that apply to objects.
     * @typeparam V return type for each rule builder method
     */
    export interface ObjectValidationReturning<V> extends InvertibleValidation {
        /**
         * Returns a new validator that checks that the validated value is an object that is an instance of the given class (or a subclass).
         *
         * @param ctor a constructor function for the (super) class
         */
        instanceOf(ctor: Function): V;
        /**
         * Returns a new validator that checks that the validated value is an object for which the keys pass the given validator.
         *
         * @param arrayValidator an array validator (e.g. `some` or `every`) to apply to the object keys
         */
        keys(arrayValidator: Validator): V;
        /**
         * Returns a new validator that checks that the validated value is an object for which the values pass the given validator.
         *
         * @param arrayValidator an array validator (e.g. `some` or `every`) to apply to the object values
         */
        values(arrayValidator: Validator): V;
        /**
         * Returns a new validator that checks that the validated value is an object with properties that pass the given validator.
         *
         * @param propertyValidator an object mapping property names to their respective validators
         * @param only whether an object with additional, unvalidated properties is valid
         */
        properties<T>(propertyValidator: {
            [P in keyof T]: Validator;
        }, only?: boolean): V;
    }
    /**
     * Mix-in type for non-inverted validation builders that apply to objects.
     */
    export type ObjectValidation = ObjectValidationReturning<ObjectValidation & Validator>;
    /**
     * Mix-in interface for building validators that validate types
     * and narrow the set of available validation methods accordingly.
     */
    export interface NarrowingTypeValidation {
        /** Returns a new validator that checks that the validated value is an array. */
        array(): ArrayValidation & Validator;
        /** Returns a new validator that checks that the validated value is a boolean value (`true` or `false`). */
        boolean(): AnyValidation & Validator;
        /** Returns a new validator that checks that the validated value is an integer. */
        integer(): NumericValidation & Validator;
        /** Returns a new validator that checks that the validated value is `null`. */
        null(): AnyValidation & Validator;
        /** Returns a new validator that checks that the validated value is a `number` (including `NaN`). */
        number(): NumericValidation & Validator;
        /** Returns a new validator that checks that the validated value is numeric (including numeric strings and Number but excluding `NaN`). */
        numeric(): NumericValidation & Validator;
        /** Returns a new validator that checks that the validated value is an object (not including arrays). */
        object(): ObjectValidation & Validator;
        /** Returns a new validator that checks that the validated value is a string. */
        string(): StringValidation & Validator;
        /** Returns a new validator that checks that the validated value is `undefined`. */
        undefined(): AnyValidation & Validator;
    }
    /**
     * Combined type for type-narrowing/non-inverted validation.
     */
    export type NarrowingValidation = NarrowingTypeValidation & NumericValidation & LengthValidation<StringValidation & ArrayValidation & Validator> & StringValidation & ArrayValidation & ObjectValidation & Validator;
    /**
     * Mix-in interface for building an inverted validator that checks that values are **not** of a given type.
     * Inverted validation is inherently non-narrowing (i.e. knowing what type something **isn't** does not allow
     * practical restriction of validation methods to the remaining possible types).
     * @typeparam V return type for each rule builder method
     */
    export interface InvertedTypeValidation<V> {
        /** Returns a new validator that checks that the validated value is an array. */
        array(): V;
        /** Returns a new validator that checks that the validated value is a boolean value (`true` or `false`). */
        boolean(): V;
        /** Returns a new validator that checks that the validated value is an integer. */
        integer(): V;
        /** Returns a new validator that checks that the validated value is `null`. */
        null(): V;
        /** Returns a new validator that checks that the validated value is a `number` (including `NaN`). */
        number(): V;
        /** Returns a new validator that checks that the validated value is numeric (including numeric strings but excluding `NaN`). */
        numeric(): V;
        /** Returns a new validator that checks that the validated value is an object (not including arrays). */
        object(): V;
        /** Returns a new validator that checks that the validated value is a string. */
        string(): V;
        /** Returns a new validator that checks that the validated value is `undefined`. */
        undefined(): V;
    }
    /**
     * Combined type for building an inverted/non-narrowing validator.
     * The type parameter represents the type of the validator before inversion,
     * so that it can be restored after the next rule is added.
     * @typeparam V validator type before inversion
     */
    export type InvertedValidation<V> = InvertedTypeValidation<V> & NumericValidationReturning<V> & StringValidationReturning<V> & ArrayValidationReturning<V> & ObjectValidationReturning<V>;
}
declare module "Validator" {
    import { RuleInstance, ValidationError } from "api";
    export class Validator {
        readonly rules: RuleInstance[];
        protected readonly allowNull: boolean;
        protected readonly allowUndefined: boolean;
        constructor(rules: RuleInstance[], allowNull: boolean, allowUndefined: boolean);
        protected withRule(rule: RuleInstance): this;
        optional(): this;
        orNull(): this;
        orUndefined(): this;
        describe(): string;
        test(value: any): boolean;
        check(value: any, name?: string): void;
        testAll(value: any, name?: string): ValidationError[];
        private getFailingRules;
        private getRuleError;
    }
}
declare module "Validation" {
    import { Predicate, RuleInstance } from "api";
    import { Validator } from "Validator";
    export class Validation extends Validator {
        private readonly invertNext;
        constructor(rules: RuleInstance[], allowNull?: boolean, allowUndefined?: boolean, invertNext?: boolean);
        protected withRule(rule: RuleInstance): this;
        array(): this;
        boolean(): this;
        integer(): this;
        null(): this;
        number(): this;
        numeric(): this;
        object(): this;
        string(): this;
        undefined(): this;
        get not(): Validation;
        anyOf(...validators: Validator[]): this;
        equal(value: any): this;
        exact(value: any): this;
        predicate(test: Predicate, description: string): this;
        between(min: number, max: number): this;
        greaterThan(bound: number): this;
        greaterThanOrEqual(bound: number): this;
        lessThan(bound: number): this;
        lessThanOrEqual(bound: number): this;
        positive(): this;
        negative(): this;
        length(min: number, max?: number): this;
        minLength(min: number): this;
        maxLength(max: number): this;
        empty(): this;
        pattern(regex: RegExp): this;
        contains(str: string): this;
        startsWith(str: string): this;
        endsWith(str: string): this;
        includes(value: any): this;
        every(elementValidator: Validator): this;
        some(elementValidator: Validator): this;
        instanceOf(ctor: Function): this;
        keys(arrayValidator: Validator): this;
        values(arrayValidator: Validator): this;
        properties<T>(propertyValidator: {
            [P in keyof T]: Validator;
        }, only?: boolean): this;
    }
}
declare module "index" {
    import { NarrowingValidation } from "api";
    export { ValidationError } from "api";
    export { Validation } from "Validation";
    export { Validator } from "Validator";
    const tsfv: NarrowingValidation;
    export default tsfv;
}
